 classdef Scott_SP_Quality_Control_Check_Toolbox
     
     properties

           MarkerSet = {'A',    'VICON_WAND'    ,'required';...
                        'B',    'VICON_WAND'    ,'required';...
                        'C',    'VICON_WAND'    ,'required';...
                        'D',    'VICON_WAND'    ,'optional';...
                        'E',    'VICON_WAND'    ,'required'};
     end
    
  methods           
  
  %% 1) FIND THE C3D FILES AND CHECK BOTH TRIALS FOR CHECK ARE PRESENT ====
  function [Selected_Static, Selected_Dynamic]= CheckCorrectTrials (~, Origin, StaticTrial_Filename, DynamicTrial_Filename)
      

  %% IDENTIFY THE C3D FILES IN THE SELECTED FOLDER ========================   
  files = searchFolder4Files(Origin,'');
  IDX = cell(size(files,1),1);
            
  for n = 1:size(files,1)
                               
      if strcmp(files{n}(end-3:end),'.c3d') && ...
         ~isempty(regexpi(files{n},StaticTrial_Filename))
                         
         IDX{n} = 'Staticc3d'; % Identify static c3d
   
                    
  elseif strcmp(files{n}(end-3:end),'.c3d') && ...
         ~isempty(regexpi(files{n},DynamicTrial_Filename))
                    
         IDX{n} = 'Dynamicc3d'; % Identify dynamic c3d
                  
     end         

  end
  
  Static   = files(strcmp(IDX,'Staticc3d')); % collate static c3d into variable "Static"
  Dynamic   = files(strcmp(IDX,'Dynamicc3d')); % collate dynamic c3d into variable "Dynamic"
  
  
  %% IF THE STATIC, DYNAMIC OR BOTH ARE MISSING, NOTIFY THE USER ==========
    
  if isempty (Static) && ~isempty(Dynamic)
         error('There is no c3d file for the static check. Please Check.'); %If no static c3d files generate alert
     
  elseif ~isempty (Static) && isempty(Dynamic)
         error('There is no c3d file for the dynamic check. Please Check.'); %If no dynamic c3d files generate alert
      
  elseif isempty (Static) && isempty(Dynamic)
         error('There is no c3d file for the static or dynamic check. Please Check.'); %If no static or dynamic c3d files generate alert
       
  end
  
  %% CHECK IF MORE THAN ONE STATIC C3D FILE IS AVAILABLE ==================

      if size(Static,1) == 1
          
        Selected_Static = char(Static);
                
  elseif nnz(strcmp(IDX,'Staticc3d'))>1 
         % have user select correct Static.c3d if multiple are found       
         s = listdlg('PromptString','Select the Static file you would like to use',...
                     'SelectionMode','single',...
                     'ListSize',[600 200],...
                     'ListString',Static);
         
         Selected_Static = char(Static{s});
                
      end
     
      
  %% CHECK IF MORE THAN ONE DYNAMIC C3D FILE IS AVAILABLE ==================
                      
  if size(Dynamic,1) == 1
      
        Selected_Dynamic = char(Dynamic);
                
  elseif nnz(strcmp(IDX,'Dynamicc3d'))>1 
         % have user select correct Dynamic.c3d if multiple are found
         s = listdlg('PromptString','Select the Dynamic file you would like to use',...
                            'SelectionMode','single',...
                            'ListSize',[600 200],...
                            'ListString',Dynamic);
         
         Selected_Dynamic = char(Dynamic{s});
             
  end
  
  
  end
  
  %% 2) CALCULATE THE EXPANDED UNCERTAINTY FOR THE STATIC TRIAL ===========
  function [Distances, Angles] = QuantifyExpandedUncertainty (control,Selected_Static, Distances, Angles)  
           
   % Open C3D File --------------------------------------------------------
     c3dfile = btkReadAcquisition(char(Selected_Static)); %open c3d file
            
   % Get marker data ------------------------------------------------------
     [marker, markerInfo] = btkGetPoints(c3dfile);
              
     
  %% Check if all markers are present based on the control marker set =====
  
     MARKERNAMES = control.MarkerSet(:,1);
     
     allmarks = fieldnames(marker);
     
     for m = 1:size(allmarks,1)
         if strlength(allmarks(m))~= 1
     marker = rmfield(marker,allmarks(m));
         end
     end
     
     
     for n = 1:size(MARKERNAMES,1)
         if sum(strcmp(fieldnames(marker),MARKERNAMES{n})) == 0 && ...
            strcmp(control.MarkerSet{n,3},'required')
            btkCloseAcquisition(c3dfile);
            error(['Essential marker(s) missing from static:(',MARKERNAMES{n},') Cannot process']);
         end
     end
     
     
  %% CALCUTE RECONSTRUCTED INTER-MARKER GEOMETRIES ========================
  
   allmarks = fieldnames(marker); % update with refined list of marker names

   for n = 1:length(marker.(allmarks {1})) 
       
      %% First, calculate the distances ===================================
      
      IntermarkerDistances = fieldnames(Distances);
      
      for m = 1:length(IntermarkerDistances)
          
      if sum(marker.(Distances.(IntermarkerDistances{m}).PrimaryMarker)(n,1:3) ~=0) && sum(marker.(Distances.(IntermarkerDistances{m}).SecondaryMarker)(n,1:3) ~=0)...
         &&(markerInfo.label.(Distances.(IntermarkerDistances{m}).PrimaryMarker)(n,1)~=-1) &&(markerInfo.label.(Distances.(IntermarkerDistances{m}).SecondaryMarker)(n,1)~=-1)
     
      Distances.(IntermarkerDistances{m}).Static.ReconstructedDistance(n,1) = pdist2(marker.(Distances.(IntermarkerDistances{m}).PrimaryMarker)(n,1:3),marker.(Distances.(IntermarkerDistances{m}).SecondaryMarker)(n,1:3)); 
      Distances.(IntermarkerDistances{m}).Static.DistanceError(n,1) = Distances.(IntermarkerDistances{m}).KnownDistance - Distances.(IntermarkerDistances{m}).Static.ReconstructedDistance(n,1); 
      
      else
          
      Distances.(IntermarkerDistances{m}).Static.ReconstructedDistance(n,1) =  NaN;
      Distances.(IntermarkerDistances{m}).Static.DistanceError(n,1) = NaN;
      
      end
      end
      
      %% Then calculate the angles ========================================
      
      IntermarkerAngles = fieldnames(Angles);
      
      for m = 1:length(IntermarkerAngles)
          
      if sum(marker.(Angles.(IntermarkerAngles{m}).OriginMarker)(n,1:3) ~=0) && sum(marker.(Angles.(IntermarkerAngles{m}).HorizontalAxisMarker)(n,1:3) ~=0) &&...
         sum(marker.(Angles.(IntermarkerAngles{m}).VerticalAxisMarker)(n,1:3) ~=0)&&(markerInfo.label.(Angles.(IntermarkerAngles{m}).OriginMarker)(n,1)~=-1) &&...
         (markerInfo.label.(Angles.(IntermarkerAngles{m}).HorizontalAxisMarker)(n,1)~=-1) && (markerInfo.label.(Angles.(IntermarkerAngles{m}).VerticalAxisMarker)(n,1)~=-1)
     
       a = pdist2(marker.(Angles.(IntermarkerAngles{m}).HorizontalAxisMarker)(n,1:3),marker.(Angles.(IntermarkerAngles{m}).VerticalAxisMarker)(n,1:3));
       b = pdist2(marker.(Angles.(IntermarkerAngles{m}).OriginMarker)(n,1:3),marker.(Angles.(IntermarkerAngles{m}).VerticalAxisMarker)(n,1:3));
       c = pdist2(marker.(Angles.(IntermarkerAngles{m}).OriginMarker)(n,1:3),marker.(Angles.(IntermarkerAngles{m}).HorizontalAxisMarker)(n,1:3));
          
       First = b^2 + c^2;
       Second = 2*b*c;
       Cosine = (((a^2)-(First))/(Second));
          
      Angles.(IntermarkerAngles{m}).Static.ReconstructedAngle(n,1) = acosd(Cosine); 
      Angles.(IntermarkerAngles{m}).Static.AngleError(n,1) = Angles.(IntermarkerAngles{m}).KnownAngle - Angles.(IntermarkerAngles{m}).Static.ReconstructedAngle(n,1); 
      
      else
          
      Angles.(IntermarkerAngles{m}).Static.ReconstructedAngle(n,1) =  NaN;
      Angles.(IntermarkerAngles{m}).Static.AngleError(n,1) = NaN;
      
      end
      end

   end
   
   
   %% CALCULATE THE EXPANDED UNCERTAINTY ==================================
   
    for m = 1:length(IntermarkerDistances)
    
    Distances.(IntermarkerDistances{m}).Static.ExpandedUncertainty = (nanstd(Distances.(IntermarkerDistances{m}).Static.DistanceError))*3;
    
    end
    
    for m = 1:length(IntermarkerAngles)
    
    Angles.(IntermarkerAngles{m}).Static.ExpandedUncertainty = (nanstd(Angles.(IntermarkerAngles{m}).Static.AngleError))*3; 
        
    end
                   
  end
  
  %% 3) CALCULATE THE RMSE FOR THE DYNAMIC TRIAL ==========================
  function [Distances, Angles] = QuantifyRMSE (control,Selected_Dynamic, Distances, Angles)  
           
   % Open C3D File --------------------------------------------------------
     c3dfile = btkReadAcquisition(char(Selected_Dynamic)); %open c3d file
            
   % Get marker data ------------------------------------------------------
     [marker, markerInfo] = btkGetPoints(c3dfile);
              
     
  %% Check if all markers are present based on the control marker set =====
  
     MARKERNAMES = control.MarkerSet(:,1);
     
     allmarks = fieldnames(marker);
     
     for m = 1:size(allmarks,1)
         if strlength(allmarks(m))~= 1
     marker = rmfield(marker,allmarks(m));
         end
     end
     
     
     for n = 1:size(MARKERNAMES,1)
         if sum(strcmp(fieldnames(marker),MARKERNAMES{n})) == 0 && ...
            strcmp(control.MarkerSet{n,3},'required')
            btkCloseAcquisition(c3dfile);
            error(['Essential marker(s) missing from static:(',MARKERNAMES{n},') Cannot process']);
         end
     end
     
     
  %% CALCUTE RECONSTRUCTED INTER-MARKER GEOMETRIES ========================
  
   allmarks = fieldnames(marker); % update with refined list of marker names

   for n = 1:length(marker.(allmarks {1})) 
       
      %% First, calculate the distances ===================================
      
      IntermarkerDistances = fieldnames(Distances);
      
      for m = 1:length(IntermarkerDistances)
          
      if sum(marker.(Distances.(IntermarkerDistances{m}).PrimaryMarker)(n,1:3) ~=0) && sum(marker.(Distances.(IntermarkerDistances{m}).SecondaryMarker)(n,1:3) ~=0)...
         &&(markerInfo.label.(Distances.(IntermarkerDistances{m}).PrimaryMarker)(n,1)~=-1) &&(markerInfo.label.(Distances.(IntermarkerDistances{m}).SecondaryMarker)(n,1)~=-1)
     
      Distances.(IntermarkerDistances{m}).Dynamic.ReconstructedDistance(n,1) = pdist2(marker.(Distances.(IntermarkerDistances{m}).PrimaryMarker)(n,1:3),marker.(Distances.(IntermarkerDistances{m}).SecondaryMarker)(n,1:3)); 
      Distances.(IntermarkerDistances{m}).Dynamic.DistanceError(n,1) = Distances.(IntermarkerDistances{m}).KnownDistance - Distances.(IntermarkerDistances{m}).Dynamic.ReconstructedDistance(n,1); 
      
      else
          
      Distances.(IntermarkerDistances{m}).Dynamic.ReconstructedDistance(n,1) =  NaN;
      Distances.(IntermarkerDistances{m}).Dynamic.DistanceError(n,1) = NaN;
      
      end
      end
      
      %% Then calculate the angles ========================================
      
      IntermarkerAngles = fieldnames(Angles);
      
      for m = 1:length(IntermarkerAngles)
          
      if sum(marker.(Angles.(IntermarkerAngles{m}).OriginMarker)(n,1:3) ~=0) && sum(marker.(Angles.(IntermarkerAngles{m}).HorizontalAxisMarker)(n,1:3) ~=0) &&...
         sum(marker.(Angles.(IntermarkerAngles{m}).VerticalAxisMarker)(n,1:3) ~=0)&&(markerInfo.label.(Angles.(IntermarkerAngles{m}).OriginMarker)(n,1)~=-1) &&...
         (markerInfo.label.(Angles.(IntermarkerAngles{m}).HorizontalAxisMarker)(n,1)~=-1) && (markerInfo.label.(Angles.(IntermarkerAngles{m}).VerticalAxisMarker)(n,1)~=-1)
     
       a = pdist2(marker.(Angles.(IntermarkerAngles{m}).HorizontalAxisMarker)(n,1:3),marker.(Angles.(IntermarkerAngles{m}).VerticalAxisMarker)(n,1:3));
       b = pdist2(marker.(Angles.(IntermarkerAngles{m}).OriginMarker)(n,1:3),marker.(Angles.(IntermarkerAngles{m}).VerticalAxisMarker)(n,1:3));
       c = pdist2(marker.(Angles.(IntermarkerAngles{m}).OriginMarker)(n,1:3),marker.(Angles.(IntermarkerAngles{m}).HorizontalAxisMarker)(n,1:3));
          
       First = b^2 + c^2;
       Second = 2*b*c;
       Cosine = (((a^2)-(First))/(Second));
          
      Angles.(IntermarkerAngles{m}).Dynamic.ReconstructedAngle(n,1) = acosd(Cosine); 
      Angles.(IntermarkerAngles{m}).Dynamic.AngleError(n,1) = Angles.(IntermarkerAngles{m}).KnownAngle - Angles.(IntermarkerAngles{m}).Dynamic.ReconstructedAngle(n,1); 
      
      else
          
      Angles.(IntermarkerAngles{m}).Dynamic.ReconstructedAngle(n,1) =  NaN;
      Angles.(IntermarkerAngles{m}).Dynamic.AngleError(n,1) = NaN;
      
      end
      end

   end
   
   
   %% CALCULATE THE RMSE ==================================================
   
    for m = 1:length(IntermarkerDistances)
    
    Distances.(IntermarkerDistances{m}).Dynamic.RMSE = sqrt(nanmean(Distances.(IntermarkerDistances{m}).Dynamic.DistanceError.^2));  
    
    end
    
    for m = 1:length(IntermarkerAngles)
    
    Angles.(IntermarkerAngles{m}).Dynamic.RMSE = sqrt(nanmean(Angles.(IntermarkerAngles{m}).Dynamic.AngleError.^2)); 
        
    end
                   
  end
  
 %% 4) Identify maximum errors and automatically generate a basic report ==
 function [Distances, Angles] = CreateQCCheckReport (~, Distances, Angles)
     
 %% GET THE MAXIMUM RANDOM AND SYSTEMATIC ERRORS FOR DISTANCES AND ANGLES RESPECTIVELY

 IntermarkerDistances = fieldnames(Distances);
 EUDist = nan(length(IntermarkerDistances),1);
 RMSEDist = nan(length(IntermarkerDistances),1);
 
 IntermarkerAngles = fieldnames(Angles);
 EUAng = nan(length(IntermarkerAngles),1);
 RMSEAng = nan(length(IntermarkerAngles),1);
     
 for m = 1:length(IntermarkerDistances)
     
     EUDist(m,1) = Distances.(IntermarkerDistances{m}).Static.ExpandedUncertainty;
     RMSEDist(m,1) = Distances.(IntermarkerDistances{m}).Dynamic.RMSE;     
 end
 
 for m = 1:length(IntermarkerAngles)
     
     EUAng(m,1) = Angles.(IntermarkerAngles{m}).Static.ExpandedUncertainty;
     RMSEAng(m,1) = Angles.(IntermarkerAngles{m}).Dynamic.RMSE;     
 end
      
 %% Generate basic report - can be ammended by user to lab/study specifications
 
  fig = figure; 
  set(fig,'Position',[150,494,500,306])
  set (fig,'Name','Report: SP Quality Control Check ',...
       'NumberTitle','off',...
       'color'      ,'w', 'Visible','on');  
   
  % Add a title to the Report ---------------------------------------------
    uicontrol('Style','text',...,...
              'String','SP Quality Control Check Report',...
              'Units' ,'Normalized',...
              'BackgroundColor','w' ,...
              'ForegroundColor','k','FontWeight','Bold','FontSize',18,...
              'Position',[0.03,0.820261437908497,0.950000000000001,0.118135997988939]);
          
          
  % Add static results to Report ------------------------------------------
    uicontrol('Style','text',...,...
              'String',['Date: ' date],...
              'Units' ,'Normalized',...
              'BackgroundColor','w' ,...
              'ForegroundColor','k','FontWeight','Bold','FontSize',12,...
              'Position',[0.03,0.738562091503268,0.950000000000001,0.089131121166415]);
          
       
  % Add subtitle for static results to Report -----------------------------
    uicontrol('Style','text',...,...
              'String','Random Noise Errors',...
              'Units' ,'Normalized',...
              'BackgroundColor','w' ,...
              'ForegroundColor',[0.00,0.45,0.74],'FontWeight','Bold','FontSize',14,...
              'Position',[0.03,0.584967320261438,0.950000000000001,0.096812468577175]);
          
          
  % Add static results to Report ------------------------------------------
    uicontrol('Style','text',...,...
              'String',['Distance: ' num2str(max(EUDist),'%.1f'), 'mm','         ','Angle: ' num2str(max(EUAng),'%.1f') ,char(176)],...
              'Units' ,'Normalized',...
              'BackgroundColor','w' ,...
              'ForegroundColor','k','FontWeight','Bold','FontSize',12,...
              'Position',[0.03,0.480392156862746,0.950000000000001,0.087934539969834]);
          
  % Add subtitle for dynamic results to Report ----------------------------
    uicontrol('Style','text',...,...
              'String','Systematic Errors',...
              'Units' ,'Normalized',...
              'BackgroundColor','w' ,...
              'ForegroundColor',[0.47,0.67,0.19],'FontWeight','Bold','FontSize',14,...
              'Position',[0.03,0.333333333333334,0.950000000000001,0.083313222724988]);
          
    uicontrol('Style','text',...,...
              'String',['Distance: ' num2str(max(RMSEDist),'%.1f'), 'mm','         ','Angle: ' num2str(max(RMSEAng),'%.1f') ,char(176)],...
              'Units' ,'Normalized',...
              'BackgroundColor','w' ,...
              'ForegroundColor','k','FontWeight','Bold','FontSize',12,...
              'Position',[0.03,0.173202614379086,0.950000000000001,0.126551030668678]);
   
  

 
 end      
   end
 end
 


    function files = searchFolder4Files(Origin,files)
        folder = dir(Origin);
        for n = 3:size(folder,1)
                suborigin = [Origin,filesep,folder(n).name];
            switch isfolder(suborigin)
              case 1
                files = searchFolder4Files(suborigin,files);
              case 0
                files{size(files,1)+1,1} = suborigin;
            end       
        end
    end
  
